// Copyright 2017 tsuru authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package pool

import (
	"reflect"

	"github.com/globalsign/mgo/bson"
	check "gopkg.in/check.v1"
)

func (s *S) TestSetPoolConstraints(c *check.C) {
	coll := s.storage.PoolsConstraints()
	err := SetPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"planb", "hipache"}})
	c.Assert(err, check.IsNil)
	err = SetPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: ConstraintTypeTeam, Values: []string{"user"}, Blacklist: true})
	c.Assert(err, check.IsNil)
	var cs []*PoolConstraint
	err = coll.Find(bson.M{"poolexpr": "*"}).All(&cs)
	c.Assert(err, check.IsNil)
	c.Assert(cs, check.DeepEquals, []*PoolConstraint{
		{PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"planb", "hipache"}},
		{PoolExpr: "*", Field: ConstraintTypeTeam, Values: []string{"user"}, Blacklist: true},
	})
}

func (s *S) TestSetPoolConstraintsWithServices(c *check.C) {
	coll := s.storage.PoolsConstraints()
	err := SetPoolConstraint(&PoolConstraint{
		PoolExpr: "prod",
		Field:    ConstraintTypeService,
		Values:   []string{"lux"},
	})
	c.Assert(err, check.IsNil)
	err = SetPoolConstraint(&PoolConstraint{
		PoolExpr: "dev",
		Field:    ConstraintTypeService,
		Values:   []string{"demacia"},
	})
	c.Assert(err, check.IsNil)
	var cs []*PoolConstraint
	err = coll.Find(bson.M{"field": "service"}).All(&cs)
	c.Assert(err, check.IsNil)
	c.Assert(cs, check.DeepEquals, []*PoolConstraint{
		{PoolExpr: "prod", Field: ConstraintTypeService, Values: []string{"lux"}},
		{PoolExpr: "dev", Field: ConstraintTypeService, Values: []string{"demacia"}},
	})
}

func (s *S) TestSetPoolConstraintsRemoveEmpty(c *check.C) {
	coll := s.storage.PoolsConstraints()
	err := SetPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"planb", "hipache"}})
	c.Assert(err, check.IsNil)
	err = SetPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: ConstraintTypeTeam, Values: []string{"user"}, Blacklist: true})
	c.Assert(err, check.IsNil)
	err = SetPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: ConstraintTypeTeam, Values: []string{""}})
	c.Assert(err, check.IsNil)
	var cs []*PoolConstraint
	err = coll.Find(bson.M{"poolexpr": "*"}).All(&cs)
	c.Assert(err, check.IsNil)
	c.Assert(cs, check.DeepEquals, []*PoolConstraint{
		{PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"planb", "hipache"}},
	})
}

func (s *S) TestSetPoolConstraintInvalidConstraintType(c *check.C) {
	coll := s.storage.PoolsConstraints()
	err := SetPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: "invalid", Values: []string{"abc"}, Blacklist: true})
	c.Assert(err, check.Equals, ErrInvalidConstraintType)
	var cs []*PoolConstraint
	err = coll.Find(bson.M{"poolexpr": "*"}).All(&cs)
	c.Assert(err, check.IsNil)
	c.Assert(len(cs), check.Equals, 0)
}

func (s *S) TestGetConstraintsForPool(c *check.C) {
	err := SetPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"planb"}})
	c.Assert(err, check.IsNil)
	err = SetPoolConstraint(&PoolConstraint{PoolExpr: "pp", Field: ConstraintTypeRouter, Values: []string{"galeb"}})
	c.Assert(err, check.IsNil)
	err = SetPoolConstraint(&PoolConstraint{PoolExpr: "*_dev", Field: ConstraintTypeRouter, Values: []string{"planb_dev"}})
	c.Assert(err, check.IsNil)
	err = SetPoolConstraint(&PoolConstraint{PoolExpr: "*_dev", Field: ConstraintTypeTeam, Values: []string{"team_pool1"}, Blacklist: true})
	c.Assert(err, check.IsNil)
	err = SetPoolConstraint(&PoolConstraint{PoolExpr: "pool1_dev", Field: ConstraintTypeTeam, Values: []string{"team_pool1"}})
	c.Assert(err, check.IsNil)
	err = SetPoolConstraint(&PoolConstraint{PoolExpr: "pool1\\x", Field: ConstraintTypeTeam, Values: []string{"team_pool1x"}})
	c.Assert(err, check.IsNil)
	err = SetPoolConstraint(&PoolConstraint{PoolExpr: "*\\xdev", Field: ConstraintTypeTeam, Values: []string{"team_xdev"}})
	c.Assert(err, check.IsNil)
	tt := []struct {
		pool     string
		expected map[poolConstraintType]*PoolConstraint
	}{
		{pool: "prod", expected: map[poolConstraintType]*PoolConstraint{
			ConstraintTypeRouter: {PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"planb"}},
		}},
		{pool: "pp", expected: map[poolConstraintType]*PoolConstraint{
			ConstraintTypeRouter: {PoolExpr: "pp", Field: ConstraintTypeRouter, Values: []string{"galeb"}},
		}},
		{pool: "pool1_dev", expected: map[poolConstraintType]*PoolConstraint{
			ConstraintTypeRouter: {PoolExpr: "*_dev", Field: ConstraintTypeRouter, Values: []string{"planb_dev"}},
			ConstraintTypeTeam:   {PoolExpr: "pool1_dev", Field: ConstraintTypeTeam, Values: []string{"team_pool1"}},
		}},
		{pool: "pool2_dev", expected: map[poolConstraintType]*PoolConstraint{
			ConstraintTypeRouter: {PoolExpr: "*_dev", Field: ConstraintTypeRouter, Values: []string{"planb_dev"}},
			ConstraintTypeTeam:   {PoolExpr: "*_dev", Field: ConstraintTypeTeam, Values: []string{"team_pool1"}, Blacklist: true},
		}},
		{pool: "pp2", expected: map[poolConstraintType]*PoolConstraint{
			ConstraintTypeRouter: {PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"planb"}},
		}},
		{pool: "pool1\\x", expected: map[poolConstraintType]*PoolConstraint{
			ConstraintTypeRouter: {PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"planb"}},
			ConstraintTypeTeam:   {PoolExpr: "pool1\\x", Field: ConstraintTypeTeam, Values: []string{"team_pool1x"}},
		}},
		{pool: "abc\\xdev", expected: map[poolConstraintType]*PoolConstraint{
			ConstraintTypeRouter: {PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"planb"}},
			ConstraintTypeTeam:   {PoolExpr: "*\\xdev", Field: ConstraintTypeTeam, Values: []string{"team_xdev"}},
		}},
	}
	for i, t := range tt {
		constraints, err := getConstraintsForPool(t.pool)
		c.Check(err, check.IsNil)
		if !reflect.DeepEqual(constraints, t.expected) {
			c.Fatalf("(%d) Expected %#+v for pool %q. Got %#+v.", i, t.expected, t.pool, constraints)
		}
	}
}

func (s *S) TestAppendPoolConstraint(c *check.C) {
	err := SetPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"planb"}, Blacklist: true})
	c.Assert(err, check.IsNil)
	err = AppendPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: ConstraintTypeRouter, Values: []string{"galeb"}})
	c.Assert(err, check.IsNil)
	err = AppendPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: ConstraintTypeService, Values: []string{"autoscale"}})
	c.Assert(err, check.IsNil)
	err = AppendPoolConstraint(&PoolConstraint{PoolExpr: "*", Field: "invalid", Values: []string{"val"}})
	c.Assert(err, check.Equals, ErrInvalidConstraintType)
	constraints, err := getConstraintsForPool("*")
	c.Assert(err, check.IsNil)
	c.Assert(constraints, check.DeepEquals, map[poolConstraintType]*PoolConstraint{
		ConstraintTypeRouter:  {Field: ConstraintTypeRouter, PoolExpr: "*", Values: []string{"planb", "galeb"}, Blacklist: true},
		ConstraintTypeService: {Field: ConstraintTypeService, PoolExpr: "*", Values: []string{"autoscale"}},
	})
}

func (s *S) TestAppendPoolConstraintNewConstraint(c *check.C) {
	err := AppendPoolConstraint(&PoolConstraint{PoolExpr: "myPool", Field: ConstraintTypeRouter, Values: []string{"galeb"}})
	c.Assert(err, check.IsNil)
	constraints, err := getConstraintsForPool("myPool")
	c.Assert(err, check.IsNil)
	c.Assert(constraints, check.DeepEquals, map[poolConstraintType]*PoolConstraint{
		ConstraintTypeRouter: {Field: ConstraintTypeRouter, PoolExpr: "myPool", Values: []string{"galeb"}},
	})
}

func (s *S) TestToConstraintType(c *check.C) {
	_, err := ToConstraintType("")
	c.Assert(err, check.Equals, ErrInvalidConstraintType)
	_, err = ToConstraintType("x")
	c.Assert(err, check.Equals, ErrInvalidConstraintType)
	ct, err := ToConstraintType("team")
	c.Assert(err, check.IsNil)
	c.Assert(ct, check.Equals, ConstraintTypeTeam)
}
